@@ -1,5 +1,5 @@
-#ifndef GEMU_H
-#define GEMU_H
+#ifndef QEMU_H
+#define QEMU_H
 
 #include "thunk.h"
 
@@ -52,8 +52,6 @@ struct vm86_saved_state {
 #ifdef TARGET_ARM
 /* FPU emulator */
 #include "nwfpe/fpa11.h"
-#undef put_user
-#undef get_user
 #endif
 
 /* NOTE: we force a big alignment so that the stack stored after is
@@ -155,16 +153,16 @@ int target_msync(unsigned long start, unsigned long len, int flags);
     int size = sizeof(*ptr);\
     switch(size) {\
     case 1:\
-        x = (typeof(*ptr))ldub(ptr);\
+        x = (typeof(*ptr))ldub((void *)ptr);\
         break;\
     case 2:\
-        x = (typeof(*ptr))lduw(ptr);\
+        x = (typeof(*ptr))lduw((void *)ptr);\
         break;\
     case 4:\
-        x = (typeof(*ptr))ldl(ptr);\
+        x = (typeof(*ptr))ldl((void *)ptr);\
         break;\
     case 8:\
-        x = (typeof(*ptr))ldq(ptr);\
+        x = (typeof(*ptr))ldq((void *)ptr);\
         break;\
     default:\
         abort();\
@@ -238,4 +236,4 @@ static inline unsigned long clear_user(void *dst, unsigned long size)
         return size;
 }
 
-#endif
+#endif /* QEMU_H */
